/*
 *	rs2character.c
 *		Copyright (C)2002, 2004 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "filesfc.h"
#include "rs2decode.h"
#include "rs2item.h"
#include "rs2character.h"

#define FILEPOS_CHRNAME_NORM	0x1dc400
#define FILEPOS_CHRNAME_SP		0x1dcc00
#define FILEPOS_CHRDATA_NORM	0x1e0cc0
#define FILEPOS_CHRDATA_SP		0x1e16c0
#define FILEPOS_CHRLVUP			0x1e2527
#define FILEPOS_CHRLVUP_ALL		0x1ea4dc
#define FILEPOS_CHRLVUP_HP		0x1e2627
#define NUM_CHR_NORM			256
#define NUM_CHR_SP				16
#define SIZEOF_CHRNAME			8
#define SIZEOF_CHRDATA_NORM		10
#define SIZEOF_CHRDATA_SP		43
#define SIZEOF_CHRTP			2

static void chr_setname(void);

static char tbl_chrname_norm[NUM_CHR_NORM][256];	/* ʃLN^̖Oi[e[u */
static char tbl_chrname_sp[NUM_CHR_SP][256];		/* LN^̖Oi[e[u */
static int flag_setchrname = 0;						/* tbl_chrname_normEtbl_chrname_spɃf[^i[Ă邩ǂ */

const char *chr_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_CHR_NORM <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setchrname)
		chr_setname();

	return tbl_chrname_norm[index];
}

const char *chr_getnamesp(int index)
{
	/* `FbN */
	if (index < 0 || NUM_CHR_SP <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setchrname)
		chr_setname();

	return tbl_chrname_sp[index];
}

static void chr_setname(void)
{
	int i;

	/* LN^̖Of[^ */
	for (i = 0; i < NUM_CHR_NORM; i++)
		*tbl_chrname_norm[i] = '\0';
	for (i = 0; i < NUM_CHR_SP; i++)
		*tbl_chrname_sp[i] = '\0';

	/* ʃLN^̖Of[^ǂݍ */
	for (i = 0; i < NUM_CHR_NORM; i++)
	{
		unsigned char buf[SIZEOF_CHRNAME];

		file_read_sfc(buf, FILEPOS_CHRNAME_NORM + SIZEOF_CHRNAME * i, SIZEOF_CHRNAME);
		dec_string(tbl_chrname_norm[i], buf, sizeof(tbl_chrname_norm[i]), SIZEOF_CHRNAME);
	}

	/* LN^̖Of[^ǂݍ */
	for (i = 0; i < NUM_CHR_SP; i++)
	{
		unsigned char buf[SIZEOF_CHRNAME];

		file_read_sfc(buf, FILEPOS_CHRNAME_SP + SIZEOF_CHRNAME * i, SIZEOF_CHRNAME);
		dec_string(tbl_chrname_sp[i], buf, sizeof(tbl_chrname_sp[i]), SIZEOF_CHRNAME);
	}

	flag_setchrname = 1;
}

chrdata_t *chr_readdata(void)
{
	int i;
	chrdata_t *buf_chrdata = NULL;

	/* ʃLN^̊ef[^i[obt@mۂ */
	if ((buf_chrdata = malloc(sizeof(chrdata_t) * NUM_CHR_NORM)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ʃLN^̊ef[^ǂݍ */
	for (i = 0; i < NUM_CHR_NORM; i++)
	{
		unsigned char buf[SIZEOF_CHRDATA_NORM];

		file_read_sfc(buf, FILEPOS_CHRDATA_NORM + SIZEOF_CHRDATA_NORM * i, SIZEOF_CHRDATA_NORM);

		/* lnf[^ */
		buf_chrdata[i].life_point = buf[ 4];

		buf_chrdata[i].strength    = 10 + (buf[ 1] & 0x0f);
		buf_chrdata[i].dexterity   = 10 + ((buf[ 1] & 0xf0) >> 4);
		buf_chrdata[i].magic_power = 10 + (buf[ 2] & 0x0f);
		buf_chrdata[i].unknown     = 10 + ((buf[ 2] & 0xf0) >> 4);
		buf_chrdata[i].speed       = 10 + (buf[ 3] & 0x0f);
		buf_chrdata[i].stamina     = 10 + ((buf[ 3] & 0xf0) >> 4);

		buf_chrdata[i].level_tech_cng = -7 + ((buf[ 8] & 0xf0) >> 4);
		buf_chrdata[i].level_mag_cng  = -7 + (buf[ 8] & 0x0f);

		buf_chrdata[i].type_tech_learn = buf[ 5];
		buf_chrdata[i].type_race       = buf[ 0] & 0x0f;

		/* tOnf[^ */
		buf_chrdata[i].flag_skill_sword = ((buf[ 6] & 0x80) != 0x00);
		buf_chrdata[i].flag_skill_axe   = ((buf[ 6] & 0x40) != 0x00);
		buf_chrdata[i].flag_skill_spear = ((buf[ 6] & 0x20) != 0x00);
		buf_chrdata[i].flag_skill_bow   = ((buf[ 6] & 0x10) != 0x00);
		buf_chrdata[i].flag_skill_punch = ((buf[ 6] & 0x08) != 0x00);

		buf_chrdata[i].flag_skill_fire  = ((buf[ 7] & 0x80) != 0x00);
		buf_chrdata[i].flag_skill_water = ((buf[ 7] & 0x40) != 0x00);
		buf_chrdata[i].flag_skill_wind  = ((buf[ 7] & 0x20) != 0x00);
		buf_chrdata[i].flag_skill_earth = ((buf[ 7] & 0x10) != 0x00);
		buf_chrdata[i].flag_skill_holy  = ((buf[ 7] & 0x08) != 0x00);
		buf_chrdata[i].flag_skill_dark  = ((buf[ 7] & 0x04) != 0x00);

		buf_chrdata[i].flag_woman  = ((buf[ 0] & 0x80) != 0x00);
		buf_chrdata[i].flag_undead = ((buf[ 0] & 0x40) != 0x00);
		buf_chrdata[i].flag_boss   = ((buf[ 0] & 0x20) != 0x00);
		buf_chrdata[i].flag_frying = ((buf[ 0] & 0x10) != 0x00);
	}

	return buf_chrdata;
}

chrdatasp_t *chr_readdatasp(void)
{
	int i, j;
	chrdatasp_t *buf_chrdatasp = NULL;

	/* LN^̊ef[^i[obt@mۂ */
	if ((buf_chrdatasp = malloc(sizeof(chrdatasp_t) * NUM_CHR_SP)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* LN^̊ef[^ǂݍ */
	for (i = 0; i < NUM_CHR_SP; i++)
	{
		unsigned char buf[SIZEOF_CHRDATA_SP];

		file_read_sfc(buf, FILEPOS_CHRDATA_SP + SIZEOF_CHRDATA_SP * i, SIZEOF_CHRDATA_SP);

		/* lnf[^ */
		buf_chrdatasp[i].life_point = buf[ 4];

		buf_chrdatasp[i].strength    = 10 + (buf[ 1] & 0x0f);
		buf_chrdatasp[i].dexterity   = 10 + ((buf[ 1] & 0xf0) >> 4);
		buf_chrdatasp[i].magic_power = 10 + (buf[ 2] & 0x0f);
		buf_chrdatasp[i].unknown     = 10 + ((buf[ 2] & 0xf0) >> 4);
		buf_chrdatasp[i].speed       = 10 + (buf[ 3] & 0x0f);
		buf_chrdatasp[i].stamina     = 10 + ((buf[ 3] & 0xf0) >> 4);

		buf_chrdatasp[i].level_sword = buf[ 6];
		buf_chrdatasp[i].level_axe   = buf[ 7];
		buf_chrdatasp[i].level_spear = buf[ 8];
		buf_chrdatasp[i].level_bow   = buf[ 9];
		buf_chrdatasp[i].level_punch = buf[10];
		buf_chrdatasp[i].level_fire  = buf[11];
		buf_chrdatasp[i].level_water = buf[12];
		buf_chrdatasp[i].level_wind  = buf[13];
		buf_chrdatasp[i].level_earth = buf[14];
		buf_chrdatasp[i].level_holy  = buf[15];
		buf_chrdatasp[i].level_dark  = buf[16];

		buf_chrdatasp[i].type_tech_learn = buf[ 5];
		buf_chrdatasp[i].type_race       = buf[ 0] & 0x0f;

		/* tOnf[^ */
		buf_chrdatasp[i].flag_frying = ((buf[ 0] & 0x10) != 0x00);
		buf_chrdatasp[i].flag_boss   = ((buf[ 0] & 0x20) != 0x00);
		buf_chrdatasp[i].flag_undead = ((buf[ 0] & 0x40) != 0x00);
		buf_chrdatasp[i].flag_woman  = ((buf[ 0] & 0x80) != 0x00);

		/* CĂZ */
		for (j = 0; j < 8; j++)
			buf_chrdatasp[i].tech_no[j] = buf[j + 17];

		/* CĂp */
		for (j = 0; j < 8; j++)
			buf_chrdatasp[i].mag_no[j] = buf[j + 25];

		/* ACe */
		for (j = 0; j < 8; j++)
			buf_chrdatasp[i].equipment_no[j] = buf[j + 33];

		/* Z|Cg̎Zo */
		{
			int max_index = 0;
			int max_level = -1;

			buf_chrdatasp[i].tech_point = 0;
			for (j = 0; j < 5; j++)
			{
				buf_chrdatasp[i].tech_point += buf[j + 6];
				if (max_level < buf[j + 6])
				{
					max_index = j;
					max_level = buf[j + 6];
				}
			}
			buf_chrdatasp[i].tech_point += buf[max_index + 6] * 2;
			if (buf_chrdatasp[i].tech_point > 250)
				buf_chrdatasp[i].tech_point = 250;
		}

		/* p|Cg̎Zo */
		{
			int max_index = 0;
			int max_level = -1;

			buf_chrdatasp[i].mag_point = 0;
			for (j = 0; j < 6; j++)
			{
				buf_chrdatasp[i].mag_point += buf[j + 11];
				if (max_level < buf[j + 11])
				{
					max_index = j;
					max_level = buf[j + 11];
				}
			}
			buf_chrdatasp[i].mag_point += buf[max_index + 11] * 2;
			if (max_level > 0)
				buf_chrdatasp[i].mag_point += 5;
			if (buf_chrdatasp[i].mag_point > 250)
				buf_chrdatasp[i].mag_point = 250;
		}
	}

	return buf_chrdatasp;
}

chrlvup_t *chr_readlvup(void)
{
	int i;
	chrlvup_t *buf_chrlvup = NULL;

	/* LN^̐f[^i[obt@mۂ */
	if ((buf_chrlvup = malloc(sizeof(chrlvup_t))) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* xƂ̕KvZp_ǂݍ */
	for (i = 0; i < NUM_CHRLV; i++)
	{
		unsigned char buf[SIZEOF_CHRTP];

		file_read_sfc(buf, FILEPOS_CHRLVUP + SIZEOF_CHRTP * i, SIZEOF_CHRTP);
		buf_chrlvup->technical_point[i] = (buf[1] << 8) | buf[0];
	}

	/* xƂ̕KvZp_ǂݍ(S̃x) */
	for (i = 0; i < NUM_CHRLV_ALL; i++)
	{
		unsigned char buf[SIZEOF_CHRTP];

		file_read_sfc(buf, FILEPOS_CHRLVUP_ALL + SIZEOF_CHRTP * i, SIZEOF_CHRTP);
		buf_chrlvup->technical_point_all[i] = (buf[1] << 8) | buf[0];
	}

	return buf_chrlvup;
}

int chr_writeparam(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ʃLN^̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ʃLN^̐lnf[^o͂ */
	fprintf(fp, "No.      O               "
				"  LP "
				"r p  s f ̗ "
				"           | ̏p "
				"           n   V   \n");
	for (i = 0; i < NUM_CHR_NORM; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %3d, %3d, %3d, %3d, %3d, %3d, %3d, %+3d, %+3d, %+3d, %+3d, %+3d, %+3d, %+3d, %+3d, %+3d, %+3d, %+3d\n", i, i,
				chr_getname(i),
				ptr_chrdata[i].life_point,
				ptr_chrdata[i].strength,
				ptr_chrdata[i].dexterity,
				ptr_chrdata[i].magic_power,
				ptr_chrdata[i].unknown,
				ptr_chrdata[i].speed,
				ptr_chrdata[i].stamina,
				(ptr_chrdata[i].flag_skill_sword) ? ptr_chrdata[i].level_tech_cng : -7,
				(ptr_chrdata[i].flag_skill_spear) ? ptr_chrdata[i].level_tech_cng : -7,
				(ptr_chrdata[i].flag_skill_axe)   ? ptr_chrdata[i].level_tech_cng : -7,
				(ptr_chrdata[i].flag_skill_bow)   ? ptr_chrdata[i].level_tech_cng : -7,
				(ptr_chrdata[i].flag_skill_punch) ? ptr_chrdata[i].level_tech_cng : -7,
				(ptr_chrdata[i].flag_skill_fire)  ? ptr_chrdata[i].level_mag_cng  : -7,
				(ptr_chrdata[i].flag_skill_water) ? ptr_chrdata[i].level_mag_cng  : -7,
				(ptr_chrdata[i].flag_skill_wind)  ? ptr_chrdata[i].level_mag_cng  : -7,
				(ptr_chrdata[i].flag_skill_earth) ? ptr_chrdata[i].level_mag_cng  : -7,
				(ptr_chrdata[i].flag_skill_holy)  ? ptr_chrdata[i].level_mag_cng  : -7,
				(ptr_chrdata[i].flag_skill_dark)  ? ptr_chrdata[i].level_mag_cng  : -7);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writeflag(const char *ptr_filename, const chrdata_t *ptr_chrdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ʃLN^̃tOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ʃLN^̃tOnf[^o͂ */
	fprintf(fp, "No.       O               "
				" s {X s 푰      MTYPE\n");
	for (i = 0; i < NUM_CHR_NORM; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s :  %2s   %2s   %2s   %2s  %-8s  TYPE%02d\n", i, i,
				chr_getname(i),
				str_getflag(ptr_chrdata[i].flag_woman),
				str_getflag(ptr_chrdata[i].flag_undead),
				str_getflag(ptr_chrdata[i].flag_boss),
				str_getflag(ptr_chrdata[i].flag_frying),
				str_getrace(ptr_chrdata[i].type_race),
				ptr_chrdata[i].type_tech_learn);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writespparam(const char *ptr_filename, const chrdatasp_t *ptr_chrdatasp)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdatasp == NULL)
	{
		assert(0);
		return 0;
	}

	/* LN^̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* LN^̐lnf[^o͂ */
	fprintf(fp, "No.      O               "
				"  LP   WP   JP "
				"r p  s f ̗ "
				"           | ̏p "
				"           n   V   \n");
	for (i = 0; i < NUM_CHR_SP; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d\n", i, i,
				chr_getnamesp(i),
				ptr_chrdatasp[i].life_point,
				ptr_chrdatasp[i].tech_point,
				ptr_chrdatasp[i].mag_point,
				ptr_chrdatasp[i].strength,
				ptr_chrdatasp[i].dexterity,
				ptr_chrdatasp[i].magic_power,
				ptr_chrdatasp[i].unknown,
				ptr_chrdatasp[i].speed,
				ptr_chrdatasp[i].stamina,
				ptr_chrdatasp[i].level_sword,
				ptr_chrdatasp[i].level_spear,
				ptr_chrdatasp[i].level_axe,
				ptr_chrdatasp[i].level_bow,
				ptr_chrdatasp[i].level_punch,
				ptr_chrdatasp[i].level_fire,
				ptr_chrdatasp[i].level_water,
				ptr_chrdatasp[i].level_wind,
				ptr_chrdatasp[i].level_earth,
				ptr_chrdatasp[i].level_holy,
				ptr_chrdatasp[i].level_dark);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writespflag(const char *ptr_filename, const chrdatasp_t *ptr_chrdatasp)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdatasp == NULL)
	{
		assert(0);
		return 0;
	}

	/* LN^̃tOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* LN^̃tOnf[^o͂ */
	fprintf(fp, "No.       O               "
				" s {X s 푰      MTYPE\n");
	for (i = 0; i < NUM_CHR_SP; i++)
	{
		fprintf(fp, "%03d(%03X): %-16s :  %2s   %2s   %2s   %2s  %-8s  TYPE%02d\n", i, i,
				chr_getnamesp(i),
				str_getflag(ptr_chrdatasp[i].flag_woman),
				str_getflag(ptr_chrdatasp[i].flag_undead),
				str_getflag(ptr_chrdatasp[i].flag_boss),
				str_getflag(ptr_chrdatasp[i].flag_frying),
				str_getrace(ptr_chrdatasp[i].type_race),
				ptr_chrdatasp[i].type_tech_learn);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writespequip(const char *ptr_filename, const chrdatasp_t *ptr_chrdatasp)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "----------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrdatasp == NULL)
	{
		assert(0);
		return 0;
	}

	/* LN^̑ACef[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* LN^̑ACef[^o͂ */
	fprintf(fp, "No.      O               "
				"1             "
				"2             "
				"3             "
				"4             "
				"h1             "
				"h2             "
				"h3             "
				"h4\n");
	for (i = 0; i < NUM_CHR_SP; i++)
	{
		const char *ptr_name[8];

		for (j = 0; j < 8; j++)
			ptr_name[j] = (ptr_chrdatasp[i].equipment_no[j] != 0xff) ? item_getname(ptr_chrdatasp[i].equipment_no[j]) : str_nothing;

		fprintf(fp, "%03d(%02X): %-16s : ", i, i, chr_getnamesp(i));
		for (j = 0; j < 8; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-16s", ptr_name[j]);
		}
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int chr_writelvup(const char *ptr_filename, const chrlvup_t *ptr_chrlvup)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_chrlvup == NULL)
	{
		assert(0);
		return 0;
	}

	/* LN^̐f[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* LN^̐f[^o͂ */
	fprintf(fp, "[Z\x]\n");
	fprintf(fp, "LV  Zp_\n");
	for (i = 0; i < NUM_CHRLV; i++)
	{
		fprintf(fp, "%2d:  %5ld\n", i + 1,
				ptr_chrlvup->technical_point[i]);
	}

	fputc('\n', fp);

	/* LN^̐f[^o͂(S̃x) */
	fprintf(fp, "[S̋Z\x]\n");
	fprintf(fp, "LV  Zp_\n");
	for (i = 0; i < NUM_CHRLV_ALL; i++)
	{
		fprintf(fp, "%2d:  %5ld\n", i + 1,
				ptr_chrlvup->technical_point_all[i]);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
