/*
 *	rs2tech.c
 *		Copyright (C)2002, 2003 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include "errmes.h"
#include "filesfc.h"
#include "rs2decode.h"
#include "rs2tech.h"

#define FILEPOS_TECHDATA		0x1e1b27
#define FILEPOS_TECHNAME		0x1dd680
#define FILEPOS_TECHLEARNFLAG	0x1e26ef
#define FILEPOS_TECHLEARNDATA	0x1e2aef
#define FILEPOS_TECHLEARNDATA2	0x02c64a
#define FILEPOS_TECHEVADEDATA	0x1e2cef
#define NUM_TECH				256
#define SIZEOF_TECHDATA			10
#define SIZEOF_TECHNAME			10
#define SIZEOF_TECHLEARNFLAG	2
#define SIZEOF_TECHLEARNDATA	2
#define SIZEOF_TECHEVADEDATA	1

static void tech_setname(void);

static char tbl_techname[NUM_TECH][256];	/* ZEp̖Oi[e[u */
static int flag_settechname = 0;			/* tbl_technameɃf[^i[Ă邩ǂ */

const char *tech_getname(int index)
{
	static const char *tbl_strname[] =
	{
		"(ʏU: )",
		"(ʏU: 匕)",
		"(ʏU: )",
		"(ʏU: _)",
		"(ʏU: )",
		"(ʏU: )",
		"(ʏU: |)",
		"(ʏU: )",
		"()",
		"()",
		"(ʏU: p`)",
	};

	assert(sizeof(tbl_strname) / sizeof(tbl_strname[0]) == 11);

	/* `FbN */
	if (index < 0 || NUM_TECH <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_settechname)
		tech_setname();

	if (0 <= index && index < 11)
		return tbl_strname[index];
	else
		return tbl_techname[index];
}

static void tech_setname(void)
{
	int i;

	/* ZEp̖Of[^ */
	for (i = 0; i < NUM_TECH; i++)
		*tbl_techname[i] = '\0';

	/* ZEp̖Of[^ǂݍ */
	for (i = 0; i < NUM_TECH; i++)
	{
		unsigned char buf[SIZEOF_TECHNAME];

		file_read_sfc(buf, FILEPOS_TECHNAME + SIZEOF_TECHNAME * i, SIZEOF_TECHNAME);
		dec_string(tbl_techname[i], buf, sizeof(tbl_techname[i]), SIZEOF_TECHNAME);
	}

	flag_settechname = 1;
}

techdata_t *tech_readdata(void)
{
	int i;
	techdata_t *buf_techdata = NULL;

	/* ZEp̊ef[^i[obt@mۂ */
	if ((buf_techdata = malloc(sizeof(techdata_t) * NUM_TECH)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* Ž؂Փxǂݍ */
	for (i = 0; i < NUM_TECH; i++)
	{
		unsigned char byte;

		file_read_sfc(&byte, FILEPOS_TECHEVADEDATA + SIZEOF_TECHEVADEDATA * i, SIZEOF_TECHEVADEDATA);
		buf_techdata[i].tech_evade_level = byte;
	}

	/* Z̑M\tOǂݍ */
	for (i = 0; i < NUM_TECH; i++)
	{
		unsigned char buf[SIZEOF_TECHLEARNFLAG];

		file_read_sfc(buf, FILEPOS_TECHLEARNFLAG + SIZEOF_TECHLEARNFLAG * i, SIZEOF_TECHLEARNFLAG);
		buf_techdata[i].bits_learn = ((unsigned long)buf[1] << 8) | buf[0];
	}

	/* ZEp̊ef[^ǂݍ */
	for (i = 0; i < NUM_TECH; i++)
	{
		unsigned char buf[SIZEOF_TECHDATA];

		file_read_sfc(buf, FILEPOS_TECHDATA + SIZEOF_TECHDATA * i, SIZEOF_TECHDATA);

		/* lnf[^ */
		buf_techdata[i].hit_rate = buf[ 0];
		buf_techdata[i].point    = buf[ 7] & 0x3f;

		/* tOnf[^ */
		buf_techdata[i].attack_elem      = buf[ 4];
		buf_techdata[i].flag_derivative  = ((buf[ 9] & 0x80) != 0x00);
		buf_techdata[i].flag_learn_evade = ((buf[ 9] & 0x40) != 0x00);
	}

	return buf_techdata;
}

techlearn_t *tech_readlearn(void)
{
	int i, j;
	unsigned char *buf1 = NULL;
	unsigned char *buf2 = NULL;
	unsigned char *ptr_buf = NULL;
	techlearn_t *buf_techlearn = NULL;
	const size_t bufsize = 1024;

	/* obt@mۂ */
	if ((buf1 = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	if ((buf2 = malloc(bufsize)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);
	memset(buf1, 0x00, bufsize);
	memset(buf2, 0x00, bufsize);
	file_read_sfc(buf1, FILEPOS_TECHLEARNDATA, bufsize);
	file_read_sfc(buf2, FILEPOS_TECHLEARNDATA2, bufsize);

	/* Z̑Mf[^i[obt@mۂ */
	if ((buf_techlearn = malloc(sizeof(techlearn_t))) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* hp^[f[^̐JEg */
	buf_techlearn->count = 0;
	ptr_buf = buf1;
	while (ptr_buf - buf1 < (ptrdiff_t)bufsize && *ptr_buf != 0xff)		/* ʏZ */
	{
		buf_techlearn->count++;
		ptr_buf += SIZEOF_TECHLEARNDATA * (ptr_buf[1] + 1);
	}
	ptr_buf = buf2;
	while (ptr_buf - buf2 < (ptrdiff_t)bufsize && *ptr_buf != 0xff)		/* JE^[Z */
	{
		buf_techlearn->count++;
		ptr_buf += SIZEOF_TECHLEARNDATA * (ptr_buf[1] + 1);
	}

	/* hp^[f[^i[obt@mۂ */
	if ((buf_techlearn->ptr_derivative = malloc(sizeof(techderivative_t) * buf_techlearn->count)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* hp^[f[^ݒ肷 */
	i = 0;
	ptr_buf = buf1;
	while (ptr_buf - buf1 < (ptrdiff_t)bufsize && *ptr_buf != 0xff)		/* ʏZ */
	{
		buf_techlearn->ptr_derivative[i].tech_no_source = ptr_buf[0];
		buf_techlearn->ptr_derivative[i].count          = ptr_buf[1];
		ptr_buf += SIZEOF_TECHLEARNDATA;

		/* MZf[^i[obt@mۂ */
		if ((buf_techlearn->ptr_derivative[i].ptr_learnlevel = malloc(sizeof(techlearnlevel_t) * buf_techlearn->ptr_derivative[i].count)) == NULL)
			err_printf(STR_FATAL_ERROR_IN_MALLOC);

		/* MZf[^ݒ肷 */
		for (j = 0; j < buf_techlearn->ptr_derivative[i].count; j++)
		{
			buf_techlearn->ptr_derivative[i].ptr_learnlevel[j].tech_no    = ptr_buf[0];
			buf_techlearn->ptr_derivative[i].ptr_learnlevel[j].tech_level = ptr_buf[1];
			ptr_buf += SIZEOF_TECHLEARNDATA;
		}

		i++;
	}

	ptr_buf = buf2;
	while (ptr_buf - buf2 < (ptrdiff_t)bufsize && *ptr_buf != 0xff)		/* JE^[Z */
	{
		buf_techlearn->ptr_derivative[i].tech_no_source = ptr_buf[0];
		buf_techlearn->ptr_derivative[i].count          = ptr_buf[1];
		ptr_buf += SIZEOF_TECHLEARNDATA;

		/* MZf[^i[obt@mۂ */
		if ((buf_techlearn->ptr_derivative[i].ptr_learnlevel = malloc(sizeof(techlearnlevel_t) * buf_techlearn->ptr_derivative[i].count)) == NULL)
			err_printf(STR_FATAL_ERROR_IN_MALLOC);

		/* MZf[^ݒ肷 */
		for (j = 0; j < buf_techlearn->ptr_derivative[i].count; j++)
		{
			buf_techlearn->ptr_derivative[i].ptr_learnlevel[j].tech_no    = ptr_buf[0];
			buf_techlearn->ptr_derivative[i].ptr_learnlevel[j].tech_level = ptr_buf[1];
			ptr_buf += SIZEOF_TECHLEARNDATA;
		}

		i++;
	}

	free(buf1);
	free(buf2);
	buf1 = NULL;
	buf2 = NULL;

	return buf_techlearn;
}

int tech_writeparam(const char *ptr_filename, const techdata_t *ptr_techdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_techdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ZEp̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ZEp̐lnf[^o͂ */
	fprintf(fp, "No.      O                   "
				" \n");
	for (i = 0; i < NUM_TECH; i++)
	{
		fprintf(fp, "%03d(%02X): %-20s :  %3d, %3d\n", i, i,
				tech_getname(i),
				ptr_techdata[i].point,
				ptr_techdata[i].hit_rate);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int tech_writelearnflag(const char *ptr_filename, const techdata_t *ptr_techdata)
{
	int i, j;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_techdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ZC̃tOnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ZC̃tOnf[^o͂ */
	fprintf(fp, "No.      O                   "
				" h TY00 TY01 TY02 TY03 TY04 TY05 TY06 TY07 TY08 TY09 TY10 TY11 TY12 TY13 TY14 TY15\n");
	for (i = 0; i < NUM_TECH; i++)
	{
		fprintf(fp, "%03d(%02X): %-20s :   %2s", i, i, tech_getname(i), str_getflag(ptr_techdata[i].flag_derivative));
		for (j = 0; j < 16; j++)
			fprintf(fp, "   %2s", str_getflag((ptr_techdata[i].bits_learn & (1U << j)) != 0));
		fputc('\n', fp);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int tech_writelearndata(const char *ptr_filename, const techlearn_t *ptr_techlearn)
{
	int i, j;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_techlearn == NULL)
	{
		assert(0);
		return 0;
	}

	/* Z̑Mf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* Z̑Mf[^o͂ */
	for (i = 0; i < ptr_techlearn->count; i++)
	{
		const int tech_no_source = ptr_techlearn->ptr_derivative[i].tech_no_source;

		if (i != 0)
			fputc('\n', fp);

		fprintf(fp, "[No.%03d(%02X): %s]\n", tech_no_source, tech_no_source, tech_getname(tech_no_source));
		fprintf(fp, "No.      O                   "
					"x\n");

		for (j = 0; j < ptr_techlearn->ptr_derivative[i].count; j++)
		{
			const int tech_no = ptr_techlearn->ptr_derivative[i].ptr_learnlevel[j].tech_no;

			fprintf(fp, "%03d(%02X): %-20s :  %3d\n", tech_no, tech_no,
					tech_getname(tech_no),
					ptr_techlearn->ptr_derivative[i].ptr_learnlevel[j].tech_level);
		}
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int tech_writeevadedata(const char *ptr_filename, const techdata_t *ptr_techdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_techdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* Ž؂Փxf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* Ž؂Փxf[^o͂ */
	fprintf(fp, "No.      O                   "
				"  x\n");
	for (i = 0; i < NUM_TECH; i++)
	{
		fprintf(fp, "%03d(%02X): %-20s :  %2s    %3d\n", i, i,
				tech_getname(i),
				str_getflag(ptr_techdata[i].flag_learn_evade),
				ptr_techdata[i].tech_evade_level);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

void tech_freelearn(techlearn_t *ptr_techlearn)
{
	int i, j;

	/* `FbN */
	if (ptr_techlearn == NULL)
	{
		assert(0);
		return;
	}

	for (i = 0; i < ptr_techlearn->count; i++)
	{
		for (j = 0; j < ptr_techlearn->ptr_derivative[i].count; j++)
		{
			free(ptr_techlearn->ptr_derivative[i].ptr_learnlevel);
			ptr_techlearn->ptr_derivative[i].ptr_learnlevel = NULL;
		}
	}

	free(ptr_techlearn->ptr_derivative);
	ptr_techlearn->ptr_derivative = NULL;
}
