/*
 *	rs3item.c
 *		Copyright (C)2002, 2003 Shingo Endo <s-endo@yk.rim.or.jp>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "errmes.h"
#include "filesfc.h"
#include "rs3decode.h"
#include "rs3tech.h"
#include "rs3item.h"

#define FILEPOS_ITEMDATA			0x3e0000
#define FILEPOS_ITEMNAME			0x3d0280
#define FILEPOS_ITEMPRICE			0x3e9958
#define NUM_ITEM_TOOL				8
#define NUM_ITEM_WEAPON				72
#define NUM_ITEM_SHIELD				16
#define NUM_ITEM_PROTECTOR			128
#define NUM_ITEM_OTHER				32
#define NUM_ITEM					(NUM_ITEM_TOOL + NUM_ITEM_WEAPON + NUM_ITEM_SHIELD + NUM_ITEM_PROTECTOR + NUM_ITEM_OTHER)	/* = 256 */
#define SIZEOF_ITEMDATA_TOOL		4
#define SIZEOF_ITEMDATA_WEAPON		4
#define SIZEOF_ITEMDATA_SHIELD		4
#define SIZEOF_ITEMDATA_PROTECTOR	9
#define SIZEOF_ITEMNAME				8
#define SIZEOF_ITEMPRICE			2
#define REGION_TOOL_BEGIN			0
#define REGION_TOOL_END				(REGION_TOOL_BEGIN + NUM_ITEM_TOOL - 1)				/* =   7 */
#define REGION_WEAPON_BEGIN			(REGION_TOOL_END + 1)								/* =   8 */
#define REGION_WEAPON_END			(REGION_WEAPON_BEGIN + NUM_ITEM_WEAPON - 1)			/* =  79 */
#define REGION_SHIELD_BEGIN			(REGION_WEAPON_END + 1)								/* =  80 */
#define REGION_SHIELD_END			(REGION_SHIELD_BEGIN + NUM_ITEM_SHIELD - 1)			/* =  95 */
#define REGION_PROTECTOR_BEGIN		(REGION_SHIELD_END + 1)								/* =  96 */
#define REGION_PROTECTOR_END		(REGION_PROTECTOR_BEGIN + NUM_ITEM_PROTECTOR - 1)	/* = 223 */
#define REGION_OTHER_BEGIN			(REGION_PROTECTOR_END + 1)							/* = 224 */
#define REGION_OTHER_END			(REGION_OTHER_BEGIN + NUM_ITEM_OTHER - 1)			/* = 255 */

static void item_setname(void);

static char tbl_itemname[NUM_ITEM][256];	/* ACe̖Oi[e[u */
static int flag_setitemname = 0;			/* tbl_itemnameɃf[^i[Ă邩ǂ */

const char *item_getname(int index)
{
	/* `FbN */
	if (index < 0 || NUM_ITEM <= index)
	{
		assert(0);
		return "";
	}

	/* e[uɒlݒ肳ĂȂꍇ́Alݒ肷 */
	if (!flag_setitemname)
		item_setname();

	return tbl_itemname[index];
}

static void item_setname(void)
{
	int i;

	/* ACe̖Of[^ */
	for (i = 0; i < NUM_ITEM; i++)
		*tbl_itemname[i] = '\0';

	/* ACe̖Of[^ǂݍ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMNAME];

		file_read_sfc(buf, FILEPOS_ITEMNAME + SIZEOF_ITEMNAME * i, SIZEOF_ITEMNAME);
		dec_string(tbl_itemname[i], buf, sizeof(tbl_itemname[i]), SIZEOF_ITEMNAME);
	}

	flag_setitemname = 1;
}

itemdata_t *item_readdata(void)
{
	int i, j;
	long filepos;
	itemdata_t *buf_itemdata = NULL;

	/* ACe̊ef[^i[obt@mۂ */
	if ((buf_itemdata = malloc(sizeof(itemdata_t) * NUM_ITEM)) == NULL)
		err_printf(STR_FATAL_ERROR_IN_MALLOC);

	/* ACẻif[^ǂݍ */
	for (i = 0; i < NUM_ITEM; i++)
	{
		unsigned char buf[SIZEOF_ITEMPRICE];

		file_read_sfc(buf, FILEPOS_ITEMPRICE + SIZEOF_ITEMPRICE * i, SIZEOF_ITEMPRICE);

		buf_itemdata[i].price = (buf[1] << 8) | buf[0];
	}

	/* ACe̊ef[^ǂݍ */
	filepos = FILEPOS_ITEMDATA;
	for (i = 0; i < NUM_ITEM; i++)
	{
		/* ACe̎ނ𔻕ʂ */
		if (REGION_TOOL_BEGIN <= i && i <= REGION_TOOL_END)
		{
			unsigned char buf[SIZEOF_ITEMDATA_TOOL];

			filepos += file_read_sfc(buf, filepos, SIZEOF_ITEMDATA_TOOL);

			buf_itemdata[i].type = ITEM_TYPE_TOOL;
		}
		else if (REGION_WEAPON_BEGIN <= i && i <= REGION_WEAPON_END)
		{
			unsigned char buf[SIZEOF_ITEMDATA_WEAPON];

			filepos += file_read_sfc(buf, filepos, SIZEOF_ITEMDATA_WEAPON);

			buf_itemdata[i].type = ITEM_TYPE_WEAPON;

			/* lnf[^ */
			buf_itemdata[i].param.weapon.attack_power = buf[ 1] & 0x7f;
			buf_itemdata[i].param.weapon.weight       = buf[ 0] & 0x07;

			/* tOnf[^ */
			buf_itemdata[i].param.weapon.flag_2handed = ((buf[ 0] & 0x80) != 0x00);
			buf_itemdata[i].param.weapon.flag_fixed   = ((buf[ 1] & 0x80) != 0x00);

			/* ŗLZf[^ */
			for (j = 0; j < 2; j++)
				buf_itemdata[i].param.weapon.tech_no[j] = buf[j + 2];
		}
		else if (REGION_SHIELD_BEGIN <= i && i <= REGION_SHIELD_END)
		{
			unsigned char buf[SIZEOF_ITEMDATA_SHIELD];

			filepos += file_read_sfc(buf, filepos, SIZEOF_ITEMDATA_SHIELD);

			buf_itemdata[i].type = ITEM_TYPE_SHIELD;

			/* lnf[^ */
			buf_itemdata[i].param.shield.defense_power = buf[ 1] & 0x7f;
			buf_itemdata[i].param.shield.evade_rate    = buf[ 2] & 0x7f;
			buf_itemdata[i].param.shield.weight        = buf[ 0] & 0x07;

			/* tOnf[^ */
			buf_itemdata[i].param.shield.regist_elem = buf[ 3];
		}
		else if (REGION_PROTECTOR_BEGIN <= i && i <= REGION_PROTECTOR_END)
		{
			unsigned char buf[SIZEOF_ITEMDATA_PROTECTOR];

			filepos += file_read_sfc(buf, filepos, SIZEOF_ITEMDATA_PROTECTOR);

			buf_itemdata[i].type = ITEM_TYPE_PROTECTOR;

			/* lnf[^ */
			buf_itemdata[i].param.protector.defense_power       = buf[ 0] & 0x7f;
			buf_itemdata[i].param.protector.magic_defense_power = buf[ 2] & 0x7f;
			buf_itemdata[i].param.protector.weight              = buf[ 8] & 0x0f;
			buf_itemdata[i].param.protector.type_healing        = buf[ 6] & 0x07;

			/* ϐf[^ */
			buf_itemdata[i].param.protector.defense_slash =
				((buf[ 1] & 0x80) != 0) ? (buf[ 1] & 0x07) : 0;
			buf_itemdata[i].param.protector.defense_knock =
				((buf[ 1] & 0x40) != 0) ? (buf[ 1] & 0x07) : 0;
			buf_itemdata[i].param.protector.defense_poke =
				((buf[ 1] & 0x20) != 0) ? (buf[ 1] & 0x07) : 0;
			buf_itemdata[i].param.protector.defense_shoot =
				((buf[ 1] & 0x10) != 0) ? (buf[ 1] & 0x07) : 0;
			buf_itemdata[i].param.protector.defense_fire =
				((buf[ 3] & 0x80) != 0) ? (buf[ 3] & 0x07) : 0;
			buf_itemdata[i].param.protector.defense_ice =
				((buf[ 3] & 0x40) != 0) ? (buf[ 3] & 0x07) : 0;
			buf_itemdata[i].param.protector.defense_thunder =
				((buf[ 3] & 0x20) != 0) ? (buf[ 3] & 0x07) : 0;
			buf_itemdata[i].param.protector.defense_state =
				((buf[ 3] & 0x10) != 0) ? (buf[ 3] & 0x07) : 0;

			/* p[^ωf[^ */
			buf_itemdata[i].param.protector.increase_strength =
				((buf[ 7] & 0x80) != 0) ? (buf[ 7] & 0x01) + 1 : 0;
			buf_itemdata[i].param.protector.increase_dexterity =
				((buf[ 7] & 0x40) != 0) ? (buf[ 7] & 0x01) + 1 : 0;
			buf_itemdata[i].param.protector.increase_speed =
				((buf[ 7] & 0x20) != 0) ? (buf[ 7] & 0x01) + 1 : 0;
			buf_itemdata[i].param.protector.increase_stamina =
				((buf[ 7] & 0x10) != 0) ? (buf[ 7] & 0x01) + 1 : 0;
			buf_itemdata[i].param.protector.increase_magic_power =
				((buf[ 7] & 0x08) != 0) ? (buf[ 7] & 0x01) + 1 : 0;
			buf_itemdata[i].param.protector.increase_will_power =
				((buf[ 7] & 0x04) != 0) ? (buf[ 7] & 0x01) + 1 : 0;
			buf_itemdata[i].param.protector.increase_charm =
				((buf[ 7] & 0x02) != 0) ? (buf[ 7] & 0x01) + 1 : 0;

			buf_itemdata[i].param.protector.regist_state = buf[ 4];

			/* tOnf[^ */
			buf_itemdata[i].param.protector.flag_regist_water     = ((buf[ 6] & 0x08) != 0x00);
			buf_itemdata[i].param.protector.flag_regist_throw     = ((buf[ 6] & 0x10) != 0x00);
			buf_itemdata[i].param.protector.flag_regist_soundwave = ((buf[ 6] & 0x20) != 0x00);
			buf_itemdata[i].param.protector.flag_regist_gaze      = ((buf[ 6] & 0x40) != 0x00);
			buf_itemdata[i].param.protector.flag_regist_ground    = ((buf[ 6] & 0x80) != 0x00);

			buf_itemdata[i].param.protector.flag_fixed = ((buf[ 5] & 0x80) != 0x00);
		}
	}

	return buf_itemdata;
}

int item_writeparam(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACe̐lnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACe̐lnf[^o͂(퓬ACe) */
	fprintf(fp, "No.      O               "
				"  i\n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %5ld\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price);
	}

	/* ACe̐lnf[^o͂() */
	fprintf(fp, "No.      O               "
				"  i "
				"U d\n");
	for (i = REGION_WEAPON_BEGIN; i <= REGION_WEAPON_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %5ld, %3d, %3d\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price,
				ptr_itemdata[i].param.weapon.attack_power,
				ptr_itemdata[i].param.weapon.weight);
	}

	/* ACe̐lnf[^o͂() */
	fprintf(fp, "No.      O               "
				"  i "
				"h  d\n");
	for (i = REGION_SHIELD_BEGIN; i <= REGION_SHIELD_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %5ld, %3d, %3d, %3d\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price,
				ptr_itemdata[i].param.shield.defense_power,
				ptr_itemdata[i].param.shield.evade_rate,
				ptr_itemdata[i].param.shield.weight);
	}

	/* ACe̐lnf[^o͂(h) */
	fprintf(fp, "No.      O               "
				"  i "
				"h h "
				"ha h h h hM h h h "
				"r p f ̗  ӎu  "
				"d\n");
	for (i = REGION_PROTECTOR_BEGIN; i <= REGION_PROTECTOR_END; i++)
	{
		const protector_t *ptr_protector = &ptr_itemdata[i].param.protector;

		fprintf(fp, "%03d(%02X): %-16s :  %5ld, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d, %3d\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price,
				ptr_protector->defense_power,
				ptr_protector->magic_defense_power,
				item_getdefense(ptr_protector->defense_power, ptr_protector->defense_slash),
				item_getdefense(ptr_protector->defense_power, ptr_protector->defense_knock),
				item_getdefense(ptr_protector->defense_power, ptr_protector->defense_poke),
				item_getdefense(ptr_protector->defense_power, ptr_protector->defense_shoot),
				item_getdefense(ptr_protector->magic_defense_power, ptr_protector->defense_fire),
				item_getdefense(ptr_protector->magic_defense_power, ptr_protector->defense_ice),
				item_getdefense(ptr_protector->magic_defense_power, ptr_protector->defense_thunder),
				item_getdefense(ptr_protector->magic_defense_power, ptr_protector->defense_state),
				ptr_protector->increase_strength,
				ptr_protector->increase_dexterity,
				ptr_protector->increase_speed,
				ptr_protector->increase_stamina,
				ptr_protector->increase_magic_power,
				ptr_protector->increase_will_power,
				ptr_protector->increase_charm,
				ptr_protector->weight);
	}

	/* ACe̐lnf[^o͂(̑) */
	fprintf(fp, "No.      O               "
				"  i\n");
	for (i = REGION_OTHER_BEGIN; i <= REGION_OTHER_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :  %5ld\n", i, i,
				item_getname(i),
				ptr_itemdata[i].price);
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeregist(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i;
	FILE *fp = NULL;
	const char *str_regist[] =
	{
		"--",
		"~",
		"",
		"",
		"+",
		"",
		"(~)",
		"",
	};

	assert(sizeof(str_regist) / sizeof(str_regist[0]) == 8);

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACȇϐnf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACȇϐnf[^o͂(퓬ACe) */
	fprintf(fp, "No.      O\n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :\n", i, i,
				item_getname(i));
	}

	/* ACȇϐnf[^o͂() */
	fprintf(fp, "No.      O\n");
	for (i = REGION_WEAPON_BEGIN; i <= REGION_WEAPON_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :\n", i, i,
				item_getname(i));
	}

	/* ACȇϐnf[^o͂() */
	fprintf(fp, "No.      O               "
				"h䑮\n");
	for (i = REGION_SHIELD_BEGIN; i <= REGION_SHIELD_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : ", i, i,
				item_getname(i));
		str_printelem(fp, ptr_itemdata[i].param.shield.regist_elem, ELEM_ALL);
		fputc('\n', fp);
	}

	/* ACȇϐnf[^o͂(h) */
	fprintf(fp, "No.      O               "
				"ha h h h hM h h h  "
				"ԕωϐ     "
				"ϐ ϓ ω ϋ ϒn\n");
	for (i = REGION_PROTECTOR_BEGIN; i <= REGION_PROTECTOR_END; i++)
	{
		if (ptr_itemdata[i].param.protector.defense_slash   >= 8 ||
			ptr_itemdata[i].param.protector.defense_knock   >= 8 ||
			ptr_itemdata[i].param.protector.defense_poke    >= 8 ||
			ptr_itemdata[i].param.protector.defense_shoot   >= 8 ||
			ptr_itemdata[i].param.protector.defense_fire    >= 8 ||
			ptr_itemdata[i].param.protector.defense_ice     >= 8 ||
			ptr_itemdata[i].param.protector.defense_thunder >= 8 ||
			ptr_itemdata[i].param.protector.defense_state   >= 8)
		{
			break;
		}

		fprintf(fp, "%03d(%02X): %-16s :", i, i,
				item_getname(i));
		fprintf(fp, "  %-4s %-4s %-4s %-4s %-4s %-4s %-4s %-4s",
				str_regist[ptr_itemdata[i].param.protector.defense_slash],
				str_regist[ptr_itemdata[i].param.protector.defense_knock],
				str_regist[ptr_itemdata[i].param.protector.defense_poke],
				str_regist[ptr_itemdata[i].param.protector.defense_shoot],
				str_regist[ptr_itemdata[i].param.protector.defense_fire],
				str_regist[ptr_itemdata[i].param.protector.defense_ice],
				str_regist[ptr_itemdata[i].param.protector.defense_thunder],
				str_regist[ptr_itemdata[i].param.protector.defense_state]);
		fprintf(fp, " ");
		str_printstate(fp, ptr_itemdata[i].param.protector.regist_state, STATE_ALL);
		fprintf(fp, "  %-2s   %-2s   %-2s   %-2s   %-2s",
				str_getflag(ptr_itemdata[i].param.protector.flag_regist_water),
				str_getflag(ptr_itemdata[i].param.protector.flag_regist_throw),
				str_getflag(ptr_itemdata[i].param.protector.flag_regist_soundwave),
				str_getflag(ptr_itemdata[i].param.protector.flag_regist_gaze),
				str_getflag(ptr_itemdata[i].param.protector.flag_regist_ground));
		fputc('\n', fp);
	}

	/* ACȇϐnf[^o͂(̑) */
	fprintf(fp, "No.      O\n");
	for (i = REGION_OTHER_BEGIN; i <= REGION_OTHER_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :\n", i, i,
				item_getname(i));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}

int item_writeeffect(const char *ptr_filename, const itemdata_t *ptr_itemdata)
{
	int i, j;
	FILE *fp = NULL;
	const char *str_nothing = "--------------------";

	/* `FbN */
	if (ptr_filename == NULL || *ptr_filename == '\0' || ptr_itemdata == NULL)
	{
		assert(0);
		return 0;
	}

	/* ACěʃf[^o͂t@CJ */
	if ((fp = fopen(ptr_filename, "w")) == NULL)
		return 0;

	/* ACěʃf[^o͂(퓬ACe) */
	fprintf(fp, "No.      O\n");
	for (i = REGION_TOOL_BEGIN; i <= REGION_TOOL_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :\n", i, i,
				item_getname(i));
	}

	/* ACěʃf[^o͂() */
	fprintf(fp, "No.      O               "
				"ŗLZ1           ŗLZ2\n");
	for (i = REGION_WEAPON_BEGIN; i <= REGION_WEAPON_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s : ", i, i,
				item_getname(i));
		for (j = 0; j < 2; j++)
		{
			if (j != 0)
				fprintf(fp, ", ");
			fprintf(fp, "%-20s",
					(ptr_itemdata[i].param.weapon.tech_no[j] != 0xff) ?
					tech_getname(ptr_itemdata[i].param.weapon.tech_no[j]) : str_nothing);
		}
		fputc('\n', fp);
	}

	/* ACěʃf[^o͂() */
	fprintf(fp, "No.      O\n");
	for (i = REGION_SHIELD_BEGIN; i <= REGION_SHIELD_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :\n", i, i,
				item_getname(i));
	}

	/* ACěʃf[^o͂(h) */
	fprintf(fp, "No.      O\n");
	for (i = REGION_PROTECTOR_BEGIN; i <= REGION_PROTECTOR_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :\n", i, i,
				item_getname(i));
	}

	/* ACěʃf[^o͂(̑) */
	fprintf(fp, "No.      O\n");
	for (i = REGION_OTHER_BEGIN; i <= REGION_OTHER_END; i++)
	{
		fprintf(fp, "%03d(%02X): %-16s :\n", i, i,
				item_getname(i));
	}

	fclose(fp);
	fp = NULL;

	return 1;
}
